const
   NULL          = 0;
   EOF           = -1;
   FLI_MAXFRAMES = 4000;
   FLIH_MAGIC    = $AF11;
   FLIF_MAGIC    = $F1FA;
   FLI_COLOR     = 11;
   FLI_LC        = 12;
   FLI_BLACK     = 13;
   FLI_BRUN      = 15;
   FLI_COPY      = 16;
   FirstFlag : boolean = FALSE;

type
   error_T = (SUCCESS,ERR_READ,ERR_WRITE,ERR_ALLOC,ERR_INV_FLIFRAME,ERR_INV_CHUNK,ERR_INV_COMP);
   buffer_T = array[0..63999] of byte;

   FLI_header_T = record
      size : longint;
      types : word;
      frame_count, width, height, BPP : word;
      flags, speed : integer;
      next_head, frames_in_table : longint;
      files : integer;
      frame1_off, strokes, session : longint;
      reserved : array[1..88] of char;
   end;

   FLI_frame_T = record
      size : longint;
      types : word;
      chunks : integer;
      pad : array[0..7] of char;
   end;

   FLI_chunk_T = record
      size : longint;
      types : word;
   end;

   palette_T = record
      re : word;
      skip, change : byte;
   end;

var
   screen_buffer : ^buffer_T;
   fCount : longint;
   SCAN_CODE : char;


function KeyPressed : boolean;
var
   flag : boolean;
begin
   flag := TRUE;
   asm
      mov ah, 1
      int $16
      jnz @KEY_PRESSED
      mov flag, 0
   @KEY_PRESSED:
   end;
   KeyPressed := flag;
end;

function readKey : char;
var
   i : integer;
begin
   if SCAN_CODE > #0 then begin
      readKey := SCAN_CODE;
      SCAN_CODE := #0;
      exit;
   end;
   asm
@MAIN_LOOP:
      mov ah, 1
      int $16
      jz  @MAIN_LOOP
      mov ah, 0
      int $16
      cmp al, 0
      jne @JUMP00
      mov SCAN_CODE, ah
@JUMP00:
      mov @RESULT, al
   end;

end;

procedure waitVerticalRetrace(count : integer);
var
   c : char;
begin
   while count > 0 do begin
      while (port[$03DA] and 8) > 0 do;
      while (port[$03DA] and 8) = 0 do;
      dec(count);
   end;
end;

function fopen(var name; mode : byte) : integer; assembler;
asm
   push ds
   lds  dx,name
   mov  ah,3dh
   mov  al,mode
   int  21h
   pop  ds
   jc   @JUMP_ERROR
   jmp  @EXIT
@JUMP_ERROR:
   mov  ax, -1;
@EXIT:
end;

function fclose(handle : integer) : integer; assembler;
asm
   mov  ah,3eh
   mov  bx,handle
   jc   @JUMP_EOF
   mov  ax, NULL;
   jmp  @EXIT
@JUMP_EOF:
   mov  ax, EOF;
@EXIT:
end;

function fread(var buf; size : word; handle : integer) : word; assembler;
asm
   push ds
   lds  dx,buf
   mov  ah,3fh
   mov  bx,handle
   mov  cx,size
   int  21h
   pop  ds
   jc   @JUMP_NULL
   jmp  @EXIT
@JUMP_NULL:
   mov  ax, NULL;
@EXIT:
end;

function fseek(handle : integer; len : longint; whence : byte) : integer; assembler;
asm
   push ds
   lds  dx,len
   mov  cx,ds
   pop  ds
   mov  ah,42h
   mov  al,whence
   mov  bx,handle
   int  21h
end;

procedure setMode(mode : byte); assembler;
asm
   mov  ah,0
   mov  al,mode
   int  10h
end;

procedure setVGAPalette(var palette; skip, change : word); assembler;
asm
   les  dx,palette
   mov  ax,1012h
   mov  bx,skip
   mov  cx,change
   int  10h
end;

function readPalette(fp : integer; size : word) : error_T;
var
   palette_header : palette_T;
   palette_data : array[0..768-1] of char;
   read_size, buffer_size : word;
   temp : integer;
begin
   read_size := fread(palette_header,sizeof(palette_T),fp);
   if (read_size <> sizeof(palette_T)) then begin
      readPalette := ERR_READ;
      exit;
   end;
   if (palette_header.skip = 0) and (palette_header.change = 0) then
      buffer_size := 256
   else
      buffer_size := palette_header.change;
   temp := fread(palette_data,size - 4,fp);
   if (temp <> size - 4) then begin
      readPalette := ERR_READ;
      exit;
   end;
   setVGAPalette(palette_data,palette_header.skip,buffer_size);
   readPalette := SUCCESS;
end;

function doRLE(fp : integer; size : word) : error_T;
label
   _END;
var
   buffer : ^buffer_T;
   count, iCount, buffer_ptr : word;
   temp_byte : byte;
   packets : integer;
begin
   iCount := 0;
   buffer_ptr := 0;
   if FirstFlag then begin
      fcount := longint(sizeof(FLI_header_T));
      move(screen_buffer^,mem[$A000:$0000],64000);
      doRLE := SUCCESS;
      exit;
   end;
   getMem(buffer,size);
   if buffer = nil then begin
      doRLE := ERR_ALLOC;
      exit;
   end;
   if (fread(buffer^,size,fp) <> size) then begin
      doRLE := ERR_READ;
      exit;
   end;
   while (iCount < 64000) do begin
      packets := word(buffer^[buffer_ptr]) and $00FF;
      inc(buffer_ptr);
      while (packets > 0) do begin
         dec(packets);
         count := word(buffer^[buffer_ptr]) and $00FF;
         inc(buffer_ptr);
         if count = 0 then goto _END
         else if count > 127 then begin
	    count := 256 - count;
            while (count > 0) do begin
               dec(count);
	       screen_buffer^[iCount] := buffer^[buffer_ptr];
               inc(iCount);
               inc(buffer_ptr);
            end;
         end else begin
	    temp_byte := buffer^[buffer_ptr];
            inc(buffer_ptr);
            while (count > 0) do begin
               dec(count);
               screen_buffer^[iCount] := temp_byte;
               inc(iCount);
            end;
         end;
      end;
   end;

_END:
   move(screen_buffer^,mem[$A000:$0000],64000);
   freeMem(buffer,size);
   FirstFlag := TRUE;
   doRLE := SUCCESS;
end;

procedure clearScreen; assembler;
asm
    mov  ax, 0A000h
    mov  es, ax
    xor  di, di
    mov  ax, 0
    mov  cx, 32000
    rep  stosw
end;

function copyToScreen(fp : integer; size : word) : error_T;
var
   buffer : ^buffer_T;
begin
   getMem(buffer,size);
   if buffer = nil then begin
      copyToScreen := ERR_ALLOC;
      exit;
   end;
   if (fread(buffer^,size,fp) <> size) then begin
      copyToScreen := ERR_READ;
      exit;
   end;
   asm
      push ds
      lds  si,buffer
      mov  ax,0A000h
      mov  es,ax
      xor  di,di
      mov  cx,size
      rep  movsb
      pop  ds
   end;
   freeMem(buffer,size);
   copyToScreen :=  SUCCESS;
end;

function doLC(fp : integer; size : word) : error_T;
label
   _LOOP00, _LOOP01, _LOOP05, _LOOP06, _LOOP07, _LOOP09;
var
   buffer : ^buffer_T;
begin
   getMem(buffer,size);
   if buffer = nil then begin
      doLC := ERR_ALLOC;
      exit;
   end;
   if (fread(buffer^,size,fp) <> size) then begin
      doLC := ERR_READ;
      exit;
   end;

   asm
      push ds
      lds  si,buffer
      mov  ax,0a000h
      mov  es,ax
      lodsw
      mov  dx,ax
      lodsw
      mov  cx,ax
      xchg dh,dl
      mov  bx,dx
      shr  bx,1
      shr  bx,1
      add  dx,bx
      mov  di,dx

_LOOP00:
      cmp  cx,0
      je   _LOOP06
      push di
      push cx
      xor  ch,ch
      lodsb
      mov  cl,al

_LOOP01:
      cmp  cl,0
      je   _LOOP05
      push cx
      xor  ah,ah
      lodsb
      add  di,ax
      xor  ch,ch
      lodsb
      mov  cl,al
      cmp  cl,7fh
      jbe  @LOOP02
      jmp  _LOOP07

@LOOP02:
      cmp  cx,0
      je   @LOOP03
      jmp  _LOOP09

@LOOP03:
      pop  cx
      pop  cx
      pop  ds
   end;

   doLC := ERR_INV_COMP;
   exit;

   asm
@LOOP04:
      pop  cx
      loop _LOOP01

_LOOP05:
      pop  cx
      pop  di
      add  di,320
      loop _LOOP00

_LOOP06:
      pop  ds
      jmp  @LOOP11

_LOOP07:
      not  cl
      inc  cx
      lodsb
      cld
      rep  stosb
      jmp  @LOOP04

@LOOP08:
      not  cl
      inc  cx
      lodsb
      mov  ah,al
      shr  cx,1
      cld
      rep  stosw
      jmp  @LOOP04

_LOOP09:
      cld
      rep  movsb
      jmp  @LOOP04

@LOOP10:
      shr  cx,1
      cld
      rep  movsw
      jmp  @LOOP04

@LOOP11:
   end;

   freeMem(buffer,size);
   doLC := SUCCESS;
end;

function doChunk(fp : integer) : error_T;
var
   FLI_chunk : FLI_chunk_T;
   read_size : word;
   return : error_T;
begin
   read_size := fread(FLI_chunk,sizeof(FLI_chunk_T),fp);
   if read_size <> sizeof(FLI_chunk_T) then begin
      doChunk := ERR_READ;
      exit;
   end;

   return := SUCCESS;
   case FLI_chunk.types of
      FLI_COLOR : return := readPalette(fp,word(FLI_chunk.size-sizeof(FLI_chunk_T)));
      FLI_LC    : return := doLC(fp,FLI_chunk.size - sizeof(FLI_chunk_T));
      FLI_BLACK : clearScreen;
      FLI_BRUN  : return := doRLE(fp,FLI_chunk.size - sizeof(FLI_chunk_T));
      FLI_COPY  : return := copyToScreen(fp,FLI_chunk.size - sizeof(FLI_chunk_T));
      else        return := ERR_INV_CHUNK;
   end;
   if return = SUCCESS then
      doChunk := SUCCESS
   else
      doChunk := return;
end;

function playOneFrame(fp : integer) : error_T;
var
   FLI_frame : FLI_frame_T;
   read_size : word;
   i : integer;
   return : error_T;
begin
   read_size := fread(FLI_frame,sizeof(FLI_frame_T),fp);
   if read_size <> sizeof(FLI_frame_T) then begin
      playOneFrame := ERR_READ;
      exit;
   end;
   if FLI_frame.types <> FLIF_MAGIC then begin
      playOneFrame := ERR_INV_FLIFRAME;
      exit;
   end;
   for i := 0 to FLI_frame.chunks - 1 do begin
      return := doChunk(fp);
      if return <> SUCCESS then begin
         playOneFrame := return;
         exit;
      end;
   end;
   inc(fCount,FLI_frame.size);
   fseek(fp,fCount,0);
   playOneFrame := SUCCESS;
end;

procedure playFLI(file_name : string);
var
   fp : integer;
   FLI_header : FLI_header_T;
   read_size : word;
   current_frame : integer;
   return : error_T;
   i, len : byte;
begin
   len := length(file_name);
   for i := 1 to len do begin
      file_name[i-1] := file_name[i];
   end;
   file_name[len] := #0;

   current_frame := 0;
   fCount := longint(sizeof(FLI_header_T));

   fp := fopen(file_name,0);
   if fp = -1 then begin
      writeLn('ERROR : Can''t find file !');
      halt(1);
   end;

   read_size := fread(FLI_header,sizeof(FLI_header_T),fp);
   if read_size <> sizeof(FLI_header_T) then begin
      writeLn('ERROR : Read Error !');
      fclose(fp);
      halt(2);
   end;

   if FLI_header.types <> FLIH_MAGIC then begin
      writeLn('ERROR : Not flic animation File !');
      fclose(fp);
      halt(1);
   end;

   getMem(screen_buffer,64000);
   if screen_buffer = nil then begin
      writeLn('ERROR : Memory allocation error !');
      fclose(fp);
      halt(1);
   end;

   setMode($13);

   return := SUCCESS;
   while (not KeyPressed) and (return = SUCCESS) do begin
      return := playOneFrame(fp);
      if current_frame >= FLI_header.frame_count then begin
         current_frame := 0;
         fseek(fp,longint(sizeof(FLI_header_T)),0);
      end;
      inc(current_frame);
      waitVerticalRetrace(FLI_header.speed);
   end;

   if return = SUCCESS then
      readKey
   else
      writeLn('Error Code : ',ord(return));

   setMode($03);
   freeMem(screen_buffer,64000);
   fclose(fp);
end;

begin
   if paramCount = 1 then begin
      playFLI(paramStr(1));
   end else begin
      writeLn('USAGE : PAS_FLI <FLI file>');
   end;
end.
